// MassBuilderSaveTool
// Copyright (C) 2021-2024 Guillaume Jacquemin
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

#include "../../BinaryIo/Reader.h"
#include "../../BinaryIo/Writer.h"
#include "../PropertySerialiser.h"
#include "../Types/IntProperty.h"
#include "../Types/NoneProperty.h"
#include "../../Logger/Logger.h"

#include "ResourceProperty.h"

using namespace Containers::Literals;

namespace Gvas { namespace Serialisers {

Types::UnrealPropertyBase::ptr
ResourceProperty::deserialiseProperty(Containers::StringView name, Containers::StringView type,
                                      std::size_t value_length, BinaryIo::Reader& reader,
                                      PropertySerialiser& serialiser)
{
    auto prop = Containers::pointer<Types::ResourceItemValue>();

    auto id_prop = serialiser.read(reader);
    if(!id_prop) {
        LOG_ERROR("Couldn't read the ID property."_s);
        return nullptr;
    }

    if((*id_prop->name) != "ID_4_AAE08F17428E229EC7A2209F51081A21"_s ||
       id_prop->propertyType != "IntProperty"_s                      ||
       dynamic_cast<Types::IntProperty*>(id_prop.get()) == nullptr)
    {
        LOG_ERROR("The ID property is invalid."_s);
        return nullptr;
    }

    prop->id = dynamic_cast<Types::IntProperty*>(id_prop.get())->value;

    auto value_prop = serialiser.read(reader);
    if(!value_prop) {
        LOG_ERROR("Couldn't read the value property."_s);
        return nullptr;
    }

    if((*value_prop->name) != "Quantity_3_560F09B5485C365D3041888910019CE3"_s ||
       value_prop->propertyType != "IntProperty"_s                            ||
       dynamic_cast<Types::IntProperty*>(value_prop.get()) == nullptr)
    {
        LOG_ERROR("The value property is invalid."_s);
        return nullptr;
    }

    prop->quantity = dynamic_cast<Types::IntProperty*>(value_prop.get())->value;

    auto none_prop = serialiser.read(reader);

    if(!none_prop                                  ||
       (*none_prop->name) != "None"_s              ||
       none_prop->propertyType != "NoneProperty"_s ||
       !dynamic_cast<Types::NoneProperty*>(none_prop.get()))
    {
        LOG_ERROR("Couldn't find a terminating NoneProperty."_s);
        return nullptr;
    }

    return prop;
}

bool
ResourceProperty::serialiseProperty(Types::UnrealPropertyBase::ptr& prop, std::size_t& bytes_written,
                                    BinaryIo::Writer& writer, PropertySerialiser& serialiser)
{
    auto res_prop = dynamic_cast<Types::ResourceItemValue*>(prop.get());
    if(!res_prop) {
        LOG_ERROR("The property is not a valid ResourceItemValue property.");
        return false;
    }

    bytes_written += writer.writeUEStringToArray("ID_4_AAE08F17428E229EC7A2209F51081A21"_s) +
                     writer.writeUEStringToArray("IntProperty"_s) +
                     writer.writeValueToArray<std::size_t>(4ull) +
                     writer.writeValueToArray<char>('\0') +
                     writer.writeValueToArray<std::int32_t>(res_prop->id);

    bytes_written += writer.writeUEStringToArray("Quantity_3_560F09B5485C365D3041888910019CE3"_s) +
                     writer.writeUEStringToArray("IntProperty"_s) +
                     writer.writeValueToArray<std::size_t>(4ull) +
                     writer.writeValueToArray<char>('\0') +
                     writer.writeValueToArray<std::int32_t>(res_prop->quantity);

    bytes_written += writer.writeUEStringToArray("None"_s);

    return true;
}

}}
