// MassBuilderSaveTool
// Copyright (C) 2021-2024 Guillaume Jacquemin
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

#include <Corrade/Containers/String.h>

#include "../../Logger/Logger.h"

#include "../../BinaryIo/Reader.h"
#include "../../BinaryIo/Writer.h"

#include "TextProperty.h"

namespace Gvas { namespace Serialisers {

Types::UnrealPropertyBase::ptr
TextProperty::deserialiseProperty(Containers::StringView name, Containers::StringView type, std::size_t value_length,
                                  BinaryIo::Reader& reader, PropertySerialiser& serialiser)
{
    auto prop = Containers::pointer<Types::TextProperty>();

    auto start_position = reader.position();

    char terminator;
    if(!reader.readChar(terminator) || terminator != '\0') {
        return nullptr;
    }

    if(reader.peekChar() > 0) {
        if(!reader.readArray(prop->flags, 8)) {
            return nullptr;
        }
    }
    else {
        if(!reader.readArray(prop->flags, 4)) {
            return nullptr;
        }
    }

    if(!reader.readChar(prop->id)) {
        return nullptr;
    }

    std::int64_t interval;

    do {
        Containers::String str;

        if(!reader.readUEString(str)) {
            return nullptr;
        }

        arrayAppend(prop->data, std::move(str));

        interval = reader.position() - start_position;
    } while(std::size_t(interval) < value_length);

    prop->value = prop->data.back();

    return prop;
}

bool
TextProperty::serialiseProperty(Types::UnrealPropertyBase::ptr& prop, std::size_t& bytes_written,
                                BinaryIo::Writer& writer, PropertySerialiser& serialiser)
{
    auto text_prop = dynamic_cast<Types::TextProperty*>(prop.get());
    if(!text_prop) {
        LOG_ERROR("The property is not a valid text property.");
        return false;
    }

    writer.writeValueToArray<char>('\0');
    bytes_written += writer.writeDataToArray<char>(text_prop->flags);
    for(const auto& str : text_prop->data) {
        bytes_written += writer.writeUEStringToArray(str);
    }

    return true;
}

}}
